import random
from typing import Dict

from controllers.ControllerAdvancedEnemyTank import ControllerAdvancedEnemyTank
from controllers.ControllerBullet import ControllerBullet
from controllers.ControllerTank import ControllerTank
from models.Game import Game
from models.GameObject import GameObject
from models.enums.EnumGameObjectDirection import EnumGameObjectDirection
from models.enums.EnumGameObjectType import EnumGameObjectType



class ControllerGame:

    _instance = None # static variable shared by all ControllerGame instances

    def __init__(self):
        if ControllerGame._instance is not None:
            raise Exception("Cannot initialize singleton twice")
        self._game: Game = None # non-static member variable
        self._controllers: Dict[GameObject, ControllerTank] = {}
        self._player: GameObject = None

    @staticmethod
    def instance():
        if ControllerGame._instance is None:
            ControllerGame._instance = ControllerGame()
        return ControllerGame._instance

    @property
    def game(self):
        return self._game

    def set_game(self, game):
        self._game = game

    def new_game(self):
        self._game = game = Game()

        for x in range(game.map_size[0]):
            for y in range(game.map_size[1]):

                # randomly choose map tile type
                game_object_options = [EnumGameObjectType.NotSet] * 50 + [
                    EnumGameObjectType.Forest,
                    EnumGameObjectType.Water,
                    EnumGameObjectType.Brick,
                    EnumGameObjectType.Steel
                ] * 3 + [EnumGameObjectType.Enemy] + [EnumGameObjectType.EnemyAdvanced]
                game_object_type = random.choice(game_object_options)

                if x == int(game.map_size[0] / 2) and y == game.map_size[1] - 1:
                    game_object_type = EnumGameObjectType.Player

                if game_object_type != EnumGameObjectType.NotSet:
                    game_object = GameObject(
                        position=[x, y],
                        game_object_type=game_object_type,
                    )
                    game.game_objects.append(game_object)

                    game_object.animation_frame_duration = random.randint(0, int(game_object.animation_frame_duration_max))
                    game_object.animation_frame_max = 1
                    if game_object_type in [EnumGameObjectType.Water, EnumGameObjectType.Enemy, EnumGameObjectType.EnemyAdvanced, EnumGameObjectType.Player]:
                        game_object.animation_frame_max = 2
                        game_object.animation_frame_duration_max = 500
                        if game_object_type not in [EnumGameObjectType.Enemy, EnumGameObjectType.EnemyAdvanced, EnumGameObjectType.Player]:
                            game_object.animation_is_animating = True
                        else:
                            game_object.animation_frame_duration_max = 200


    def update(self, delta_time: float):
        game = self._game
        for game_object in game.game_objects:
            if game_object not in self._controllers:
                if game_object.game_object_type in [EnumGameObjectType.Enemy, EnumGameObjectType.EnemyAdvanced, EnumGameObjectType.Player]:
                    self._controllers[game_object] = ControllerTank(game, game_object)
                    if game_object.game_object_type == EnumGameObjectType.Player:
                        self._player = game_object
                elif game_object.game_object_type == EnumGameObjectType.Bullet:
                    self._controllers[game_object] = ControllerBullet(game, game_object)

            if game_object in self._controllers:
                self._controllers[game_object].update(delta_time)

            if game_object.animation_is_animating:
                game_object.animation_frame_duration += delta_time
                if game_object.animation_frame_duration >= game_object.animation_frame_duration_max:
                    game_object.animation_frame_duration = 0
                    game_object.animation_frame += 1
                    if game_object.animation_frame >= game_object.animation_frame_max:
                        game_object.animation_frame = 0

    def player_fire(self):
        self._controllers[self._player].fire()

    def player_set_direction(self, direction: EnumGameObjectDirection):
        self._controllers[self._player].set_direction(direction)